let major_version = 0;;
let minor_version = 6;;

let glob (pattern: string): Str.regexp = (
	let length = String.length pattern in
	let re = Buffer.create (length * 2) in
	Buffer.add_char re '^';
	let in_bracket = ref false in
	for i = 0 to length - 1 do
		if !in_bracket then (
			begin match pattern.[i] with
			| ']' ->
				in_bracket := false;
				Buffer.add_char re ']'
			| c ->
				Buffer.add_char re c
			end
		) else (
			begin match pattern.[i] with
			| '?' ->
				Buffer.add_string re "."
			| '*' ->
				Buffer.add_string re ".*"
			| ('.' | '+' | '\\' | '^' | '$' as c) ->
				Buffer.add_char re '\\';
				Buffer.add_char re c
			| '[' ->
				in_bracket := true;
				Buffer.add_char re '['
			| c ->
				Buffer.add_char re c
			end
		)
	done;
	Buffer.add_char re '$';
	Str.regexp (Buffer.contents re)
);;

let include_trailing_path_delimiter path = (
	if path.[String.length path - 1] = '/' then path else
	path ^ "/"
);;

let rec skip_space s i = (
	if String.length s = i then i else
	match s.[i] with
	| ' ' -> skip_space s (i + 1)
	| _ -> i
);;

let rec find_space s i = (
	if String.length s = i then i else
	match s.[i] with
	| ' ' -> i
	| _ -> find_space s (i + 1)
);;

let parse_month (s: string): int = (
	begin match (String.lowercase s) with
	| "jan" -> 0
	| "feb" -> 1
	| "mar" -> 2
	| "apr" -> 3
	| "may" -> 4
	| "jun" -> 5
	| "jul" -> 6
	| "aug" -> 7
	| "sep" -> 8
	| "oct" -> 9
	| "nov" -> 10
	| "dec" -> 11
	| _ ->
		try int_of_string s - 1 with
		| Failure _ -> failwith ("parse_month:" ^ s)
	end
);;

let remove_username_and_password (url: string): string = (
	if not (String.contains url '@') then url else
	let before = (
		let before = String.sub url 0 6 in
		if before = "ftp://" then before else
		let before = String.sub url 0 4 in
		assert (before = "ftp:");
		before
	) in
	let next = (
		let p = String.index url '@' + 1 in
		String.sub url p (String.length url - p)
	) in
	before ^ next
);;

let current_time = Unix.time ();;
let current_local_time = Unix.localtime current_time;;

let parse_ftp_time (s: string) (i: int): float * int = (
	let ms = i in
	let me = find_space s ms in
	let ds = skip_space s me in
	let de = find_space s ds in
	let ts = skip_space s de in
	let te = find_space s ts in
	let t_length = te - ts in
	let tz = String.sub s ts t_length in
	let tk = String.contains tz ':' in
	if tk && t_length <> 5 && t_length <> 8 then failwith ("internal_parse_time:" ^ s);
	if tk && tz.[2] <> ':' then failwith ("internal_parse_time:" ^ s);
	if tk && t_length = 8 && tz.[5] <> ':' then failwith ("internal_parse_time:" ^ s);
	if not tk && t_length <> 4 then failwith ("internal_parse_time:" ^ s);
	let mon = parse_month (String.sub s ms (me - ms)) in
	let year = (
		if tk then (
			if current_local_time.Unix.tm_mon >= mon then (
				current_local_time.Unix.tm_year
			) else (
				current_local_time.Unix.tm_year - 1
			)
		) else (
			int_of_string tz - 1900
		)
	) in
	let t = {
		Unix.tm_year = year;
		Unix.tm_mon = mon;
		Unix.tm_mday = int_of_string (String.sub s ds (de - ds));
		Unix.tm_hour = if tk then int_of_string (String.sub tz 0 2) else 0;
		Unix.tm_min = if tk then int_of_string (String.sub tz 3 2) else 0;
		Unix.tm_sec = if tk && t_length = 8 then int_of_string (String.sub tz 6 2) else 0;
		Unix.tm_wday = 0; Unix.tm_yday = 0; Unix.tm_isdst = false}
	in
	let time, _ = Unix.mktime t in
	time, skip_space s te
);;

let parse_time (s: string): float = (
	if String.contains s '-' then (
		(* YYYY-MM-DD HH:MM:SS form *)
		if String.length s <> 19
			|| s.[4] <> '-'
			|| s.[7] <> '-'
			|| s.[10] <> ' '
			|| s.[13] <> ':'
			|| s.[16] <> ':'
		then failwith ("parse_time:" ^ s);
		let t = {
			Unix.tm_year = int_of_string (String.sub s 0 4) - 1900;
			Unix.tm_mon = int_of_string (String.sub s 5 2) - 1;
			Unix.tm_mday = int_of_string (String.sub s 8 2);
			Unix.tm_hour = int_of_string (String.sub s 11 2);
			Unix.tm_min = int_of_string (String.sub s 14 2);
			Unix.tm_sec = int_of_string (String.sub s 17 2);
			Unix.tm_wday = 0; Unix.tm_yday = 0; Unix.tm_isdst = false}
		in
		let time, _ = Unix.mktime t in
		time
	) else (
		(* FTP form *)
		let result, j = parse_ftp_time s 0 in
		if j = String.length s then result else
		failwith ("parse_time:" ^ s)
	)
);;

let format_string_of_int ?(z = 1) ?(w = 1) (n: int): string = (
	let result = (
		let rec loop b c n = (
			if n = 0 then Buffer.contents b else (
				if c > 0 && c mod 3 = 0 then Buffer.add_char b ',';
				Buffer.add_char b (char_of_int (int_of_char '0' + n mod 10));
				loop b (c + 1) (n / 10)
			)
		) in
		let rev = loop (Buffer.create w) 0 n in
		let len = String.length rev in
		let result = String.create len in
		for i = 0 to len - 1 do
			result.[i] <- rev.[len - 1 - i]
		done;
		result
	) in
	let len = String.length result in
	let result = (
		if len < z then String.make (z - len) '0' ^ result else result
	) in
	let len = String.length result in
	if len < w then String.make (w - len) '_' ^ result else result
);;

let output_time f ?(y4 = false) ?(sec = false) (t: float): unit = (
	let t = Unix.localtime t in
	if y4 then (
		output_string f (string_of_int (t.Unix.tm_year + 1900))
	) else (
		output_string f (format_string_of_int ~z:2 (t.Unix.tm_year mod 100))
	);
	output_char f '-';
	output_string f (format_string_of_int ~z:2 (t.Unix.tm_mon + 1));
	output_char f '-';
	output_string f (format_string_of_int ~z:2 (t.Unix.tm_mday));
	output_char f ' ';
	output_string f (format_string_of_int ~z:2 (t.Unix.tm_hour));
	output_char f ':';
	output_string f (format_string_of_int ~z:2 (t.Unix.tm_min));
	if sec then (
		output_char f ':';
		output_string f (format_string_of_int ~z:2 (t.Unix.tm_sec))
	)
);;

let print_time = output_time stdout;;

let start s sub = (
	let sub_length = String.length sub in
	String.length s >= sub_length && String.sub s 0 sub_length = sub
);;

let temp_ext = ".$$$";;
let ini_filename = ".ftpversa";;

type options = {
	mutable local: string;
	mutable remote: string;
	mutable ignore: Str.regexp list;
	mutable since: float;
	mutable retry: int;
	mutable dry_run: bool;
	mutable verbose: bool;
	mutable help: bool;
	mutable version: bool};;

let options: options = (
	let options: options = {
		local = "";
		remote = "";
		ignore = [
			glob ("*" ^ temp_ext);
			glob ".ftpversa";
			glob ".bzr";
			glob ".hg";
			glob ".hgignore";
			glob ".svn";
			glob ".DS_Store"];
		since = 0.0;
		retry = 2;
		dry_run = false;
		verbose = false;
		help = false;
		version = false}
	in
	let rec loop i = (
		if i >= Array.length Sys.argv then () else
		let s = Sys.argv.(i) in
		let length = String.length s in
		if s = "-help" || s = "--help" then (
			options.help <- true;
			loop (i + 1)
		) else if start s "--ignore=" then (
			let pattern = String.sub s 9 (length - 9) in
			options.ignore <- glob pattern :: options.ignore;
			loop (i + 1)
		) else if s = "--ignore" then (
			options.ignore <- glob Sys.argv.(i + 1) :: options.ignore;
			loop (i + 2)
		) else if s = "-n" || s = "--dry-run" then (
			options.dry_run <- true;
			loop (i + 1)
		) else if start s "--retry=" then (
			let count = String.sub s 8 (length - 8) in
			options.retry <- int_of_string count;
			loop (i + 1)
		) else if s = "--retry" then (
			options.retry <- int_of_string Sys.argv.(i + 1);
			loop (i + 2)
		) else if start s "--since=" then (
			let time = String.sub s 8 (length - 8) in
			options.since <- parse_time time;
			loop (i + 1)
		) else if s = "--since" then (
			options.since <- parse_time Sys.argv.(i + 1);
			loop (i + 2)
		) else if s = "-v" || s = "--verbose" then (
			options.verbose <- true;
			loop (i + 1)
		) else if s = "--version" then (
			options.version <- true;
			loop (i + 1)
		) else if s.[0] = '-' then (
			failwith ("ftpversa: illegal option " ^ s)		
		) else if length >= 4 && String.sub s 0 4 = "ftp:" then (
			options.remote <- include_trailing_path_delimiter s;
			loop (i + 1)
		) else (
			options.local <- Filename.concat s "";
			loop (i + 1)
		)
	) in
	loop 1;
	options
);;

if options.version then (
	print_int major_version;
	print_char '.';
	print_int minor_version;
	print_newline ();
	exit 0
);;

if options.help then (
	print_string "usage: ftpversa local remote\n";
	print_string "     --help        show this message.\n";
	print_string "     --ignore=glob file name to be ignored.\n";
	print_string "  -n --dry-run     do'nt actually run any operations.\n";
	print_string "     --retry=n     retry n times.\n";
	print_string "     --since=time  check files modified after the time.\n";
	print_string "  -v --verbose     print more information.\n";
	print_string "     --version     print version and exit.\n";
	exit 0
);;

if options.local <> "" then (
	let ini = Filename.concat options.local ini_filename in
	if Sys.file_exists ini then (
		let f = open_in ini in
		let (_: string) = input_line f in
		let previous_url = include_trailing_path_delimiter (input_line f) in
		close_in f;
		if options.remote = "" then (
			options.remote <- previous_url
		) else if remove_username_and_password options.remote <> previous_url && options.since = 0.0 then (
			print_string "ftpversa: remote-url be changed, please add --since option.\n";
			exit 1
		)
	)
);;

if options.local = "" || options.remote = "" then (
	print_string "ftpversa: please, tell local and/or remote dir.\n";
	exit 1
);;

let get_since (path: string): float = (
	if options.since <> 0.0 then options.since else
	begin try
		let ini = open_in (Filename.concat (Filename.concat options.local path)
			ini_filename) in
		let result = parse_time (input_line ini) in
		if options.verbose then (
			print_char ' ';
			if path <> "" then (
				print_string path;
				print_char ' '
			);
			print_string "since ";
			print_time ~y4:true result;
			print_newline ()
		);	
		close_in ini;
		result
	with
	| Sys_error _ ->
		if options.verbose then (
			print_char ' ';
			if path <> "" then (
				print_string path;
				print_string " has "
			);
			print_string "no previous state.";
			print_newline ()
		);
		0.0
	end
);;

let write_now (path: string): unit = (
	let ini = open_out (Filename.concat (Filename.concat options.local path)
		ini_filename) in
	output_time ini ~y4:true ~sec:true (Unix.time ());
	output_char ini '\n';
	let remote_path = options.remote ^ path in
	output_string ini (remove_username_and_password remote_path);
	output_char ini '\n';
	close_out ini
);;

at_exit Curl.global_cleanup;;
Curl.global_init Curl.CURLINIT_GLOBALALL;;

let curl_session (f: Curl.t -> 'a): 'a = ( 
	let errorBuffer = ref "" in
	let connection = Curl.init () in
	begin try
		Curl.set_errorbuffer connection errorBuffer;
		Curl.set_followlocation connection true;
		let result = f connection in
		Curl.cleanup connection;
		result
	with
	| Curl.CurlException _ as exn ->
		flush stdout;
		prerr_newline ();
		prerr_string "error: ";
		prerr_string !errorBuffer;
		prerr_newline ();
		prerr_string "url: ";
		prerr_string (Curl.get_effectiveurl connection);
		prerr_newline ();
		flush stderr;
		Curl.cleanup connection;
		raise exn
	end
);;

let print_stat (s: Unix.stats): unit = (
	print_char (if s.Unix.st_kind = Unix.S_DIR then 'd' else '-');
	print_char (if s.Unix.st_perm land 0o400 <> 0 then 'r' else '-');
	print_char (if s.Unix.st_perm land 0o200 <> 0 then 'w' else '-');
	print_char (if s.Unix.st_perm land 0o100 <> 0 then 'x' else '-');
	print_char (if s.Unix.st_perm land 0o040 <> 0 then 'r' else '-');
	print_char (if s.Unix.st_perm land 0o020 <> 0 then 'w' else '-');
	print_char (if s.Unix.st_perm land 0o010 <> 0 then 'x' else '-');
	print_char (if s.Unix.st_perm land 0o004 <> 0 then 'r' else '-');
	print_char (if s.Unix.st_perm land 0o002 <> 0 then 'w' else '-');
	print_char (if s.Unix.st_perm land 0o001 <> 0 then 'x' else '-');
	print_char ' ';
	print_string (format_string_of_int ~w:9 s.Unix.st_size);
	print_char ' ';
	print_time s.Unix.st_mtime
);;

let octed_perm_of_int (perm: int): string = (
	let result = "___" in
	result.[0] <- char_of_int (int_of_char '0' + (perm land 0o700) / 64);
	result.[1] <- char_of_int (int_of_char '0' + (perm land 0o070) / 8);
	result.[2] <- char_of_int (int_of_char '0' + perm land 0o007);
	result;
);;

type dir_entry = string * Unix.stats * dir_entry_type
and dir_entry_type = [
	| `file
	| `directory of dir_entry list];;

let compare_dir_entry_by_name e1 e2: int = (
	let (n1, _, _) = e1 in
	let (n2, _, _) = e2 in
	String.compare n1 n2
);;

if options.verbose then print_string "**** scanning local...\n";;

let ftp_escape s = (
	if not (String.contains s '%') then s else
	let length = String.length s in
	let b = Buffer.create (length * 3) in
	for i = 0 to length - 1 do
		match s.[i] with
		| '%' -> Buffer.add_string b "%25"
		| _ -> Buffer.add_char b s.[i]
	done;
	Buffer.contents b
);;

let rec scan_dir (level: int) (path: string): dir_entry list = (
	let rec loop h rs = (
		begin try
			let name = Unix.readdir h in
			if name = "." || name = ".." then loop h rs else
			let fullname = Filename.concat path name in
			let s = Unix.stat fullname in
			if options.verbose then (
				print_string (String.make level ' ');
				print_stat s;
				print_char ' ';
				print_string name;
				print_newline ()
			);
			let t = (
				if s.Unix.st_kind <> Unix.S_DIR then `file else (
					`directory (scan_dir (level + 1) fullname)
				)
			) in
			loop h ((name, s, t) :: rs)
		with End_of_file ->
			Unix.closedir h;
			rs
		end
	) in
	let h = Unix.opendir path in
	List.sort compare_dir_entry_by_name (loop h [])
);;

let local_tree = scan_dir 1 options.local;;

if options.verbose then print_string "**** scanning remote...\n";;

let parse_dir (s: string): (string * Unix.stats) list = (
	(* drwxr-xr-x  21 panathenaia  users   512 Apr 19 14:32 www *)
	let rec loop s rs = (
		if s = "" then List.rev rs else
		let eol = String.index s  '\n' in
		let nl = eol + 1 in
		let line = String.sub s 0 (if s.[eol - 1] = '\r' then eol - 1 else eol) in
		let rest = (String.sub s nl (String.length s - nl)) in
		(* Printf.printf "%s\n" line; *)
		if line.[0] = 't' then loop rest rs else (
			let kind = (
				if line.[0] = 'd' then Unix.S_DIR else Unix.S_REG
			) in
			let perm = (
				(if line.[1] = 'r' then 0o400 else 0o000) lor
				(if line.[2] = 'w' then 0o200 else 0o000) lor
				(if line.[3] = 'x' then 0o100 else 0o000) lor
				(if line.[4] = 'r' then 0o040 else 0o000) lor
				(if line.[5] = 'w' then 0o020 else 0o000) lor
				(if line.[6] = 'x' then 0o010 else 0o000) lor
				(if line.[7] = 'r' then 0o004 else 0o000) lor
				(if line.[8] = 'w' then 0o002 else 0o000) lor
				(if line.[9] = 'x' then 0o001 else 0o000)
			) in
			let i = skip_space line 10 in
			let i = find_space line i in
			let i = skip_space line i in
			let i = find_space line i in
			let i = skip_space line i in
			let i = find_space line i in
			let i = skip_space line i in
			let size, i = (
				let n = find_space line i in
				let s = String.sub line i (n - i) in
				int_of_string s, skip_space line n
			) in
			let time, i = parse_ftp_time line i in
			if line.[i - 1] <> ' ' then failwith "parse_dir";
			let name = (
				String.sub line i (String.length line - i)
			) in
			if name = "." || name = ".." then loop rest rs else
			let item = name, {Unix.st_kind = kind; Unix.st_size = size;
				Unix.st_dev = 0; Unix.st_ino = 0; Unix.st_rdev = 0;
				Unix.st_perm = perm; Unix.st_atime = time;
				Unix.st_ctime = time; Unix.st_mtime = time;
				Unix.st_nlink = 0; Unix.st_uid =0; Unix.st_gid = 0} in
			loop rest (item :: rs)
		)
	) in
	loop s []
);;

let rec scan_remote_dir level url connection: dir_entry list = (
	let writer accum data = (
		Buffer.add_string accum data;
		String.length data
	) in
	let result = Buffer.create 16384 in
	Curl.set_writefunction connection (writer result);
	Curl.set_url connection url;
	Curl.perform connection;
	let data = Buffer.contents result in 
	let dirs = parse_dir data in
	let result = List.map (fun (name, ({Unix.st_kind = kind} as st)) -> 
		if options.verbose then (
			print_string (String.make level ' ');
			print_stat st;
			print_char ' ';
			print_string name;
			print_newline ()
		);
		if kind <> Unix.S_DIR then (name, st, `file) else (
			let es = scan_remote_dir (level + 1) (url ^ name ^ "/") connection in
			name, st, `directory es
		)
	) dirs in
	List.sort compare_dir_entry_by_name result
);;

let remote_tree = curl_session (scan_remote_dir 1 options.remote);;

if options.verbose then print_string "**** planning...\n";;

type side = [`local | `remote];;

type action_kind = [`copy | `remove];;

type action = string * action_kind * side * string * int * dir_entry;;

let check_glob (patterns: Str.regexp list) (name: string): bool = (
	List.exists (fun p -> Str.string_match p name 0) patterns
);;

let print_ignore name = (
	print_string " ignore ";
	print_string name;
	print_newline ()
);;

let plan (root_since: float)
	(local: dir_entry list)
	(remote: dir_entry list): action list =
(
	let rec loop path since ls rs results = (
		let do_only ls rs side item results = (
			let (name, st, _) = item in
			let fullname = path ^ name in
			let results = (
				if check_glob options.ignore name then (
					if options.verbose then print_ignore fullname;
					results
				) else if st.Unix.st_mtime > since then (
					(path, `copy, side, fullname, -1, item) :: results
				) else (
					(path, `remove, side, fullname, -1, item) :: results
				)
			) in
			loop path since ls rs results
		) in
		let do_sync ls rs litem ritem results = (
			let (name, lst, lk) = litem in
			let fullname = path ^ name in
			if check_glob options.ignore name then (
				if options.verbose then print_ignore fullname;
				loop path since ls rs results
			) else (
				let (_, rst, rk) = ritem in
				begin match lk, rk with
				| `directory ltree, `directory rtree ->
					let c_since = get_since fullname in
					let results = loop (fullname ^ "/") c_since ltree rtree results in
					loop path since ls rs results
				| `file, `directory _ | `directory _, `file ->
					print_string "warning: ";
					print_string fullname;
					print_string " conflicted directory/file!";
					print_newline ();
					loop path since ls rs results
				| `file, `file ->
					let results = (
						if lst.Unix.st_size = rst.Unix.st_size then (
							let ltm = Unix.localtime lst.Unix.st_mtime in
							let rtm = Unix.localtime rst.Unix.st_mtime in
							if ltm.Unix.tm_year < current_local_time.Unix.tm_year &&
								rtm.Unix.tm_year < current_local_time.Unix.tm_year
							then (
								results
							) else if lst.Unix.st_mtime < rst.Unix.st_mtime then (
								(path, `copy, `remote, fullname, -1, ritem) :: results
							) else if lst.Unix.st_mtime > max since (rst.Unix.st_mtime +. 60.0) then (
								(path, `copy, `local, fullname, rst.Unix.st_perm, litem) :: results
							) else (
								results
							)
						) else if lst.Unix.st_mtime < rst.Unix.st_mtime then (
							(path, `copy, `remote, fullname, -1, ritem) :: results
						) else (
							(path, `copy, `local, fullname, rst.Unix.st_perm, litem) :: results
						)
					) in
					loop path since ls rs results
				end
			)
		) in
		begin match ls, rs with
		| [], [] -> results
		| l :: lr, [] -> do_only lr [] `local l results
		| [], r :: rr -> do_only [] rr `remote r results
		| l :: lr, r :: rr ->
			let c = compare_dir_entry_by_name l r in
			if c < 0 then do_only lr rs `local l results else
			if c > 0 then do_only ls rr `remote r results else
			do_sync lr rr l r results
		end
	) in
	List.rev (loop "" root_since local remote [])
);;

let actions = (
	let root_since = get_since "" in
	plan root_since local_tree remote_tree
);;

if options.verbose || options.dry_run then (
	List.iter (fun (_, action, side, name, _, entry) ->
		begin match action with
		| `copy ->
			begin match side with
			| `local -> print_string " upload "
			| `remote -> print_string " download "
			end;
		| `remove ->
			print_string " remove ";
			begin match side with
			| `local -> print_string options.local
			| `remote -> print_string options.remote
			end;
		end;
		print_string name;
		begin match entry with
		| _, _, `file -> ()
		| _, _, `directory _ -> print_string "/*"
		end;
		print_newline ()
	) actions
);;

if options.dry_run then exit 0;;

if options.verbose then print_string "**** executing...\n";;

let touch filename timestamp = (
(*
	let t = Unix.localtime timestamp in
	let cmd = Printf.sprintf "touch -m -t %.4d%.2d%.2d%.2d%.2d.%.2d %s"
		(t.Unix.tm_year + 1900) (t.Unix.tm_mon + 1) t.Unix.tm_mday
		t.Unix.tm_hour t.Unix.tm_min t.Unix.tm_sec filename
	in
	if Sys.command cmd <> 0 then failwith cmd
*)
	Unix_ex.utimes filename ~actime:timestamp ~modtime:timestamp
);;

let download ~(remote: string) ~(local: string) (name: string)
	(mtime: float) connection: unit =
(
	print_string " download ";
	print_string name;
	print_string " ...";
	let rec try_loop n = (
		try
			let temp, output = Filename.open_temp_file
				~mode:[Open_creat; Open_trunc; Open_binary] "ftpversa" ""
			in
			let writer data = (
				output_string output data;
				String.length data
			) in
			Curl.set_writefunction connection writer;
			Curl.set_url connection (remote ^ ftp_escape name);
			Curl.perform connection;
			close_out output;
			Unix.rename temp (local ^ name)
		with
		| Curl.CurlException(Curl.CURLE_RECV_ERROR, _, _) as exn ->
			if n = 0 then raise exn else (
				print_string "...";
				try_loop (n - 1) 
			)
	) in
	try_loop options.retry;
	touch (local ^ name) mtime;
	print_string "ok";
	print_newline ()
);;

let upload ~(local: string) ~(remote: string) (name: string)
	(perm: int) : unit =
(
	print_string " upload ";
	print_string name;
	print_string " ...";
	curl_session (fun connection ->
		let input = open_in_bin (local ^ name) in
		let reader bytes = (
			let s = String.create bytes in
			let read = Pervasives.input input s 0 bytes in
			if read = 0 then "" else String.sub s 0 read
		) in
		Curl.set_upload connection true;
		Curl.set_readfunction connection reader;
		Curl.set_url connection (remote ^ ftp_escape name ^ temp_ext);
		let basename = Filename.basename name in
		let tempname = basename ^ temp_ext in
		let postquote = ("rnfr " ^ tempname) :: ("rnto " ^ basename) :: [] in
		let postquote = (
			if perm < 0 then postquote else
			("site chmod " ^ octed_perm_of_int perm ^ " " ^ tempname) :: postquote
		) in
		Curl.set_postquote connection postquote;
		Curl.perform connection;
		close_in input
	);
	touch (local ^ name) (Unix.time ());
	print_string "ok";
	print_newline ()
);;

let mkd ~(remote: string) (name: string): unit = (
	print_string " make directory ";
	print_string (remote ^ name);
	print_string " ...";
	curl_session (fun connection ->
		Curl.set_writefunction connection (fun data -> String.length data);
		Curl.set_url connection remote;
		Curl.set_postquote connection ["mkd " ^ name];
		Curl.perform connection
	);
	print_string "ok";
	print_newline ()
);;

let mkdir ~(local: string) (name: string): unit = (
	print_string " make directory ";
	print_string (local ^ name);
	print_string " ...";
	Unix.mkdir (local ^ name) 0o755;
	print_string "ok";
	print_newline ()
);;

let rec download_tree ~(remote: string) ~(local: string) (name: string)
	(fs: dir_entry list) connection: unit =
(
	mkdir ~local name;
	List.iter (fun (e_name, e_st, e_kind) ->
		let r_name = name ^ "/" ^ e_name in
		if check_glob options.ignore e_name then (
			if options.verbose then print_ignore r_name
		) else (
			begin match e_kind with
			| `file ->
				download ~remote ~local r_name e_st.Unix.st_mtime connection
			| `directory cfs ->
				download_tree ~local ~remote r_name cfs connection
			end
		)
	) fs
);;

let rec upload_tree ~(local: string) ~(remote: string) (name: string)
	(fs: dir_entry list): unit =
(
	mkd ~remote:remote name;
	List.iter (fun (e_name, _, e_kind) ->
		let r_name = name ^ "/" ^ e_name in
		if check_glob options.ignore e_name then (
			if options.verbose then print_ignore r_name
		) else (
			begin match e_kind with
			| `file ->
				upload ~local ~remote r_name ~-1
			| `directory cfs ->
				upload_tree ~local ~remote r_name cfs
			end
		)
	) fs
);;

let dele ~(remote: string) (name: string) connection: unit = (
	print_string " remove ";
	print_string (remote ^ name);
	print_string " ...";
	Curl.set_writefunction connection (fun data -> String.length data);
	Curl.set_url connection remote;
	Curl.set_postquote connection ["dele " ^ name];
	Curl.perform connection;
	print_string "ok";
	print_newline ()
);;

let rm ~(local: string) (name: string): unit = (
	print_string " remove ";
	print_string (local ^ name);
	print_string " ...";
	Unix.unlink (local ^ name);
	print_string "ok";
	print_newline ()	
);;

let rmd ~(remote: string) (name: string) connection: unit = (
	print_string " remove directory ";
	print_string (remote ^ name);
	print_string " ...";
	Curl.set_writefunction connection (fun data -> String.length data);
	Curl.set_url connection remote;
	Curl.set_postquote connection ["rmd " ^ name];
	Curl.perform connection;
	print_string "ok";
	print_newline ()
);;

let rmdir ~(local: string) (name: string): unit = (
	print_string " remove directory ";
	print_string (local ^ name);
	print_string " ...";
	Unix.rmdir (local ^ name);
	print_string "ok";
	print_newline ()
);;

let rec rm_tree ~(local: string) (name: string) fs: unit = (
	List.iter (fun (e_name, _, e_kind) ->
		let r_name = name ^ "/" ^ e_name in
		begin match e_kind with
		| `file ->
			rm ~local r_name
		| `directory cfs ->
			rm_tree ~local r_name cfs
		end
	) fs;
	rmdir ~local name
);;

let rec rmd_tree ~(remote: string) (name: string) fs connection: unit = (
	List.iter (fun (e_name, _, e_kind) ->
		let r_name = name ^ "/" ^ e_name in
		begin match e_kind with
		| `file ->
			dele ~remote r_name connection
		| `directory cfs ->
			rmd_tree ~remote r_name cfs connection
		end
	) fs;
	rmd ~remote name connection
);;

List.iter (fun (path, action, side, name, perm, entry) ->
	let _, e_st, e_kind = entry in
	begin match action with
	| `copy ->
		begin match side with
		| `local ->
			begin match e_kind with
			| `file ->
				upload ~local:options.local ~remote:options.remote name perm
			| `directory fs ->
				upload_tree ~local:options.local ~remote:options.remote name fs
			end;
		| `remote ->
			begin match e_kind with
			| `file ->
				curl_session (download
					~remote:options.remote ~local:options.local name e_st.Unix.st_mtime)
			| `directory fs ->
				curl_session (download_tree
					~remote:options.remote ~local:options.local name fs)
			end;
		end
	| `remove ->
		begin match side with
		| `local ->
			begin match e_kind with
			| `file ->
				rm ~local:options.local name
			| `directory fs ->
				rm_tree ~local:options.local name fs
			end;
		| `remote ->
			begin match e_kind with
			| `file ->
				curl_session (dele ~remote:options.remote name)
			| `directory fs ->
				curl_session (rmd_tree ~remote:options.remote name fs)
			end;
		end
	end;
	write_now path
) actions;;

let rec write_now_if_no_record (path: string): unit = (
	let rec loop h = (
		begin try
			let name = Unix.readdir h in
			if name = "." || name = ".." then loop h else
			let fullname = Filename.concat path name in
			let s = Unix.stat (Filename.concat options.local fullname) in
			if s.Unix.st_kind = Unix.S_DIR then (
				write_now_if_no_record fullname
			);
			loop h
		with End_of_file ->
			Unix.closedir h
		end
	) in
	write_now path;
	let h = Unix.opendir (Filename.concat options.local path) in
	loop h
);;

write_now_if_no_record "";;

